@@ -19,16 +19,16 @@ limitations under the License.
 package e2e
 
 import (
+	"errors"
 	"fmt"
 	"math"
 	"net/http"
 	"strconv"
 	"strings"
-	"sync/atomic"
 	"testing"
 	"time"
 
-	"github.com/pkg/errors"
+	"github.com/davecgh/go-spew/spew"
 	vegeta "github.com/tsenart/vegeta/lib"
 	"golang.org/x/sync/errgroup"
 	"knative.dev/pkg/system"
@@ -45,7 +45,6 @@ import (
 	v1a1test "knative.dev/serving/test/v1alpha1"
 
 	corev1 "k8s.io/api/core/v1"
-	v1 "k8s.io/api/core/v1"
 	"k8s.io/apimachinery/pkg/api/resource"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/util/wait"
@@ -53,11 +52,12 @@ import (
 )
 
 const (
-	autoscaleExpectedOutput = "399989"
 	// Concurrency must be high enough to avoid the problems with sampling
 	// but not high enough to generate scheduling problems.
-	containerConcurrency = 6
+	containerConcurrency = 6.0
 	targetUtilization    = 0.7
+	successRateSLO       = 0.999
+	autoscaleSleep       = 500
 )
 
 type testContext struct {
@@ -65,18 +65,16 @@ type testContext struct {
 	clients           *test.Clients
 	names             test.ResourceNames
 	resources         *v1a1test.ResourceObjects
-	deploymentName    string
-	domain            string
 	targetUtilization float64
-	targetValue       int
+	targetValue       float64
 	metric            string
 }
 
 func getVegetaTarget(kubeClientset *kubernetes.Clientset, domain, endpointOverride string, resolvable bool) (vegeta.Target, error) {
 	if resolvable {
 		return vegeta.Target{
-			Method: "GET",
-			URL:    fmt.Sprintf("http://%s?sleep=100", domain),
+			Method: http.MethodGet,
+			URL:    fmt.Sprintf("http://%s?sleep=%d", domain, autoscaleSleep),
 		}, nil
 	}
 
@@ -94,96 +92,44 @@ func getVegetaTarget(kubeClientset *kubernetes.Clientset, domain, endpointOverri
 	h := http.Header{}
 	h.Set("Host", domain)
 	return vegeta.Target{
-		Method: "GET",
-		URL:    fmt.Sprintf("http://%s?sleep=100", endpoint),
+		Method: http.MethodGet,
+		URL:    fmt.Sprintf("http://%s?sleep=%d", endpoint, autoscaleSleep),
 		Header: h,
 	}, nil
 }
 
-// TODO: use vegeta to send traffic for this func as well.
-func generateTraffic(ctx *testContext, concurrency int, duration time.Duration, stopChan chan struct{}) error {
-	var (
-		totalRequests      int32
-		successfulRequests int32
-		group              errgroup.Group
-	)
+func generateTraffic(
+	ctx *testContext,
+	attacker *vegeta.Attacker,
+	pacer vegeta.Pacer,
+	duration time.Duration,
+	stopChan chan struct{}) error {
 
-	ctx.t.Logf("Maintaining %d concurrent requests for %v.", concurrency, duration)
-	client, err := pkgTest.NewSpoofingClient(ctx.clients.KubeClient, ctx.t.Logf, ctx.domain, test.ServingFlags.ResolvableDomain)
-	if err != nil {
-		return fmt.Errorf("error creating spoofing client: %v", err)
-	}
-	for i := 0; i < concurrency; i++ {
-		group.Go(func() error {
-			req, err := http.NewRequest(http.MethodGet, fmt.Sprintf("http://%s?sleep=100", ctx.domain), nil)
-			if err != nil {
-				return fmt.Errorf("error creating HTTP request: %v", err)
-			}
-			done := time.After(duration)
-			for {
-				select {
-				case <-stopChan:
-					ctx.t.Log("Stopping generateTraffic")
-					return nil
-				case <-done:
-					ctx.t.Log("Time is up; done")
-					return nil
-				default:
-					atomic.AddInt32(&totalRequests, 1)
-					res, err := client.Do(req)
-					if err != nil {
-						ctx.t.Logf("Error making request %v", err)
-						continue
-					}
-
-					if res.StatusCode != http.StatusOK {
-						ctx.t.Logf("Status = %d, want: 200", res.StatusCode)
-						ctx.t.Logf("Response: %s", res)
-						continue
-					}
-					atomic.AddInt32(&successfulRequests, 1)
-				}
-			}
-		})
-	}
-
-	ctx.t.Log("Waiting for all requests to complete.")
-	if err := group.Wait(); err != nil {
-		return fmt.Errorf("error making requests for scale up: %v", err)
-	}
-
-	if successfulRequests != totalRequests {
-		return fmt.Errorf("error making requests for scale up. total = %d, errors = %d, expected 0",
-			totalRequests, totalRequests-successfulRequests)
-	}
-	return nil
-}
-
-func generateTrafficAtFixedRPS(ctx *testContext, rps int, duration time.Duration, stopChan chan struct{}) error {
-	var (
-		totalRequests      int32
-		successfulRequests int32
-	)
-
-	rate := vegeta.Rate{Freq: rps, Per: time.Second}
 	target, err := getVegetaTarget(
-		ctx.clients.KubeClient.Kube, ctx.domain, pkgTest.Flags.IngressEndpoint, test.ServingFlags.ResolvableDomain)
+		ctx.clients.KubeClient.Kube, ctx.resources.Route.Status.URL.URL().Hostname(), pkgTest.Flags.IngressEndpoint, test.ServingFlags.ResolvableDomain)
 	if err != nil {
-		return fmt.Errorf("error creating vegeta target: %v", err)
+		return fmt.Errorf("error creating vegeta target: %w", err)
 	}
-	targeter := vegeta.NewStaticTargeter(target)
-	attacker := vegeta.NewAttacker(vegeta.Timeout(duration))
 
-	ctx.t.Logf("Maintaining %v RPS requests for %v.", rps, duration)
-
-	// Start the attack!
-	results := attacker.Attack(targeter, rate, duration, "load-test")
+	results := attacker.Attack(vegeta.NewStaticTargeter(target), pacer, duration, "load-test")
 	defer attacker.Stop()
 
+	var (
+		totalRequests      int32
+		successfulRequests int32
+	)
 	for {
 		select {
 		case <-stopChan:
 			ctx.t.Log("Stopping generateTraffic")
+			successRate := float64(1)
+			if totalRequests > 0 {
+				successRate = float64(successfulRequests) / float64(totalRequests)
+			}
+			if successRate < successRateSLO {
+				return fmt.Errorf("request success rate under SLO: total = %d, errors = %d, rate = %f, SLO = %f",
+					totalRequests, totalRequests-successfulRequests, successRate, successRateSLO)
+			}
 			return nil
 		case res, ok := <-results:
 			if !ok {
@@ -194,26 +140,36 @@ func generateTrafficAtFixedRPS(ctx *testContext, rps int, duration time.Duration
 			totalRequests++
 			if res.Code != http.StatusOK {
 				ctx.t.Logf("Status = %d, want: 200", res.Code)
-				ctx.t.Logf("Response: %v", res)
+				ctx.t.Log("Response:\n" + spew.Sprint(res))
 				continue
 			}
 			successfulRequests++
 		}
 	}
+}
 
-	if successfulRequests != totalRequests {
-		return fmt.Errorf("error making requests for scale up. total = %d, errors = %d, expected 0",
-			totalRequests, totalRequests-successfulRequests)
-	}
-	return nil
+func generateTrafficAtFixedConcurrency(ctx *testContext, concurrency int, duration time.Duration, stopChan chan struct{}) error {
+	pacer := vegeta.ConstantPacer{} // Sends requests as quickly as possible, capped by MaxWorkers below.
+	attacker := vegeta.NewAttacker(vegeta.Timeout(duration), vegeta.Workers(uint64(concurrency)), vegeta.MaxWorkers(uint64(concurrency)))
+
+	ctx.t.Logf("Maintaining %d concurrent requests for %v.", concurrency, duration)
+	return generateTraffic(ctx, attacker, pacer, duration, stopChan)
+}
+
+func generateTrafficAtFixedRPS(ctx *testContext, rps int, duration time.Duration, stopChan chan struct{}) error {
+	pacer := vegeta.ConstantPacer{Freq: rps, Per: time.Second}
+	attacker := vegeta.NewAttacker(vegeta.Timeout(duration))
+
+	ctx.t.Logf("Maintaining %v RPS requests for %v.", rps, duration)
+	return generateTraffic(ctx, attacker, pacer, duration, stopChan)
 }
 
 // setup creates a new service, with given service options.
-// It returns a testContext that has resources, K8s clients,
-// and the deployment.
+// It returns a testContext that has resources, K8s clients and other needed
+// data points.
 // It sets up CleanupOnInterrupt as well that will destroy the resources
 // when the test terminates.
-func setup(t *testing.T, class, metric string, target int, targetUtilization float64, fopts ...rtesting.ServiceOption) *testContext {
+func setup(t *testing.T, class, metric string, target float64, targetUtilization float64, fopts ...rtesting.ServiceOption) *testContext {
 	t.Helper()
 	clients := Setup(t)
 
@@ -223,35 +179,39 @@ func setup(t *testing.T, class, metric string, target int, targetUtilization flo
 		Image:   "autoscale",
 	}
 	test.CleanupOnInterrupt(func() { test.TearDown(clients, names) })
-	resources, err := v1a1test.CreateRunLatestServiceReady(t, clients, &names,
-		append(fopts, rtesting.WithConfigAnnotations(map[string]string{
-			autoscaling.ClassAnnotationKey:             class,
-			autoscaling.MetricAnnotationKey:            metric,
-			autoscaling.TargetAnnotationKey:            strconv.FormatFloat(float64(target), 'f', -1, 64),
-			autoscaling.TargetUtilizationPercentageKey: strconv.FormatFloat(targetUtilization*100, 'f', -1, 64),
-		}), rtesting.WithResourceRequirements(corev1.ResourceRequirements{
-			Limits: corev1.ResourceList{
-				corev1.ResourceMemory: resource.MustParse("512Mi"),
-			},
-			Requests: corev1.ResourceList{
-				corev1.ResourceMemory: resource.MustParse("300Mi"),
-			},
-		}))...)
+	resources, _, err := v1a1test.CreateRunLatestServiceReady(t, clients, &names,
+		false, /* https TODO(taragu) turn this on after helloworld test running with https */
+		append([]rtesting.ServiceOption{
+			rtesting.WithConfigAnnotations(map[string]string{
+				autoscaling.ClassAnnotationKey:             class,
+				autoscaling.MetricAnnotationKey:            metric,
+				autoscaling.TargetAnnotationKey:            strconv.FormatFloat(target, 'f', -1, 64),
+				autoscaling.TargetUtilizationPercentageKey: strconv.FormatFloat(targetUtilization*100, 'f', -1, 64),
+				// We run the test for 60s, so make window a bit shorter,
+				// so that we're operating in sustained mode and the pod actions stopped happening.
+				autoscaling.WindowAnnotationKey: "50s",
+			}), rtesting.WithResourceRequirements(corev1.ResourceRequirements{
+				Limits: corev1.ResourceList{
+					corev1.ResourceMemory: resource.MustParse("512Mi"),
+				},
+				Requests: corev1.ResourceList{
+					corev1.ResourceMemory: resource.MustParse("300Mi"),
+				},
+			}),
+		}, fopts...)...)
 	if err != nil {
 		t.Fatalf("Failed to create initial Service: %v: %v", names.Service, err)
 	}
 
-	domain := resources.Route.Status.URL.Host
+	url := resources.Route.Status.URL.URL()
 	if _, err := pkgTest.WaitForEndpointState(
 		clients.KubeClient,
 		t.Logf,
-		domain,
-		// Istio doesn't expose a status for us here: https://github.com/istio/istio/issues/6082
-		// TODO(tcnghia): Remove this when https://github.com/istio/istio/issues/882 is fixed.
+		url,
 		v1a1test.RetryingRouteInconsistency(pkgTest.MatchesAllOf(pkgTest.IsStatusOK)),
 		"CheckingEndpointAfterUpdating",
 		test.ServingFlags.ResolvableDomain); err != nil {
-		t.Fatalf("Error probing domain %s: %v", domain, err)
+		t.Fatalf("Error probing %s: %v", url, err)
 	}
 
 	return &testContext{
@@ -259,8 +219,6 @@ func setup(t *testing.T, class, metric string, target int, targetUtilization flo
 		clients:           clients,
 		names:             names,
 		resources:         resources,
-		deploymentName:    resourcenames.Deployment(resources.Revision),
-		domain:            domain,
 		targetUtilization: targetUtilization,
 		targetValue:       target,
 		metric:            metric,
@@ -268,8 +226,9 @@ func setup(t *testing.T, class, metric string, target int, targetUtilization flo
 }
 
 func assertScaleDown(ctx *testContext) {
-	if err := WaitForScaleToZero(ctx.t, ctx.deploymentName, ctx.clients); err != nil {
-		ctx.t.Fatalf("Unable to observe the Deployment named %s scaling down: %v", ctx.deploymentName, err)
+	deploymentName := resourcenames.Deployment(ctx.resources.Revision)
+	if err := WaitForScaleToZero(ctx.t, deploymentName, ctx.clients); err != nil {
+		ctx.t.Fatalf("Unable to observe the Deployment named %s scaling down: %v", deploymentName, err)
 	}
 
 	// Account for the case where scaling up uses all available pods.
@@ -277,9 +236,9 @@ func assertScaleDown(ctx *testContext) {
 
 	if err := pkgTest.WaitForPodListState(
 		ctx.clients.KubeClient,
-		func(p *v1.PodList) (bool, error) {
+		func(p *corev1.PodList) (bool, error) {
 			for _, pod := range p.Items {
-				if strings.Contains(pod.Name, ctx.deploymentName) &&
+				if strings.Contains(pod.Name, deploymentName) &&
 					!strings.Contains(pod.Status.Reason, "Evicted") {
 					return false, nil
 				}
@@ -287,7 +246,7 @@ func assertScaleDown(ctx *testContext) {
 			return true, nil
 		},
 		"WaitForAvailablePods", test.ServingNamespace); err != nil {
-		ctx.t.Fatalf("Waiting for Pod.List to have no non-Evicted pods of %q: %v", ctx.deploymentName, err)
+		ctx.t.Fatalf("Waiting for Pod.List to have no non-Evicted pods of %q: %v", deploymentName, err)
 	}
 
 	ctx.t.Log("The Revision should remain ready after scaling to zero.")
@@ -298,15 +257,28 @@ func assertScaleDown(ctx *testContext) {
 	ctx.t.Log("Scaled down.")
 }
 
-func numberOfPods(ctx *testContext) (int32, error) {
-	deployment, err := ctx.clients.KubeClient.Kube.Apps().Deployments(test.ServingNamespace).Get(ctx.deploymentName, metav1.GetOptions{})
+func numberOfPods(ctx *testContext) (float64, error) {
+	// SKS name matches that of revision.
+	n := ctx.resources.Revision.Name
+	sks, err := ctx.clients.NetworkingClient.ServerlessServices.Get(n, metav1.GetOptions{})
 	if err != nil {
-		return 0, errors.Wrapf(err, "Failed to get deployment %q", deployment)
+		ctx.t.Logf("Error getting SKS %q: %v", n, err)
+		return 0, fmt.Errorf("error retrieving sks %q: %w", n, err)
 	}
-	return deployment.Status.ReadyReplicas, nil
+	if sks.Status.PrivateServiceName == "" {
+		ctx.t.Logf("SKS %s has not yet reconciled", n)
+		// Not an error, but no pods either.
+		return 0, nil
+	}
+	eps, err := ctx.clients.KubeClient.Kube.CoreV1().Endpoints(test.ServingNamespace).Get(
+		sks.Status.PrivateServiceName, metav1.GetOptions{})
+	if err != nil {
+		return 0, fmt.Errorf("failed to get endpoints %s: %w", sks.Status.PrivateServiceName, err)
+	}
+	return float64(resources.ReadyAddressCount(eps)), nil
 }
 
-func assertAutoscaleUpToNumPods(ctx *testContext, curPods, targetPods int32, duration time.Duration, quick bool) {
+func assertAutoscaleUpToNumPods(ctx *testContext, curPods, targetPods float64, duration time.Duration, quick bool) {
 	ctx.t.Helper()
 	// There are two test modes: quick, and not quick.
 	// 1) Quick mode: succeeds when the number of pods meets targetPods.
@@ -316,21 +288,19 @@ func assertAutoscaleUpToNumPods(ctx *testContext, curPods, targetPods int32, dur
 	// Relax the bounds to reduce the flakiness caused by sampling in the autoscaling algorithm.
 	// Also adjust the values by the target utilization values.
 
-	minPods := int32(math.Floor(float64(curPods)/ctx.targetUtilization)) - 1
-	maxPods := int32(math.Ceil(float64(targetPods)/ctx.targetUtilization)) + 1
+	minPods := math.Floor(curPods/ctx.targetUtilization) - 1
+	maxPods := math.Ceil(targetPods/ctx.targetUtilization) + 1
 
 	stopChan := make(chan struct{})
 	var grp errgroup.Group
-	switch ctx.metric {
-	case autoscaling.RPS:
-		grp.Go(func() error {
-			return generateTrafficAtFixedRPS(ctx, int(targetPods)*ctx.targetValue, duration, stopChan)
-		})
-	default:
-		grp.Go(func() error {
-			return generateTraffic(ctx, int(targetPods)*ctx.targetValue, duration, stopChan)
-		})
-	}
+	grp.Go(func() error {
+		switch ctx.metric {
+		case autoscaling.RPS:
+			return generateTrafficAtFixedRPS(ctx, int(targetPods*ctx.targetValue), duration, stopChan)
+		default:
+			return generateTrafficAtFixedConcurrency(ctx, int(targetPods*ctx.targetValue), duration, stopChan)
+		}
+	})
 
 	grp.Go(func() error {
 		// Short-circuit traffic generation once we exit from the check logic.
@@ -347,7 +317,7 @@ func assertAutoscaleUpToNumPods(ctx *testContext, curPods, targetPods int32, dur
 				if err != nil {
 					return err
 				}
-				mes := fmt.Sprintf("deployment '%s' #replicas: %d, want at least: %d", ctx.deploymentName, got, minPods)
+				mes := fmt.Sprintf("revision %q #replicas: %v, want at least: %v", ctx.resources.Revision.Name, got, minPods)
 				ctx.t.Log(mes)
 				if got < minPods {
 					return errors.New(mes)
@@ -356,13 +326,13 @@ func assertAutoscaleUpToNumPods(ctx *testContext, curPods, targetPods int32, dur
 					// A quick test succeeds when the number of pods scales up to `targetPods`
 					// (and, for sanity check, no more than `maxPods`).
 					if got >= targetPods && got <= maxPods {
-						ctx.t.Logf("got %d replicas, reached target of %d, exiting early", got, targetPods)
+						ctx.t.Logf("Got %v replicas, reached target of %v, exiting early", got, targetPods)
 						return nil
 					}
 				}
 				if minPods < targetPods-1 {
 					// Increase `minPods`, but leave room to reduce flakiness.
-					minPods = int32(math.Min(float64(got), float64(targetPods))) - 1
+					minPods = math.Min(got, targetPods) - 1
 				}
 			case <-done:
 				// The test duration is over. Do a last check to verify that the number of pods is at `targetPods`
@@ -371,7 +341,8 @@ func assertAutoscaleUpToNumPods(ctx *testContext, curPods, targetPods int32, dur
 				if err != nil {
 					return err
 				}
-				mes := fmt.Sprintf("got %d replicas, expected between [%d, %d] replicas for deployment %s", got, targetPods-1, maxPods, ctx.deploymentName)
+				mes := fmt.Sprintf("got %v replicas, expected between [%v, %v] replicas for revision %s",
+					got, targetPods-1, maxPods, ctx.resources.Revision.Name)
 				ctx.t.Log(mes)
 				if got < targetPods-1 || got > maxPods {
 					return errors.New(mes)
@@ -410,7 +381,7 @@ func TestAutoscaleUpCountPods(t *testing.T) {
 		name, class := name, class
 		t.Run(name, func(tt *testing.T) {
 			tt.Parallel()
-			cancel := logstream.Start(t)
+			cancel := logstream.Start(tt)
 			defer cancel()
 
 			ctx := setup(tt, class, autoscaling.Concurrency, containerConcurrency, targetUtilization)
@@ -509,17 +480,17 @@ func TestTargetBurstCapacity(t *testing.T) {
 	const duration = time.Hour
 
 	grp.Go(func() error {
-		return generateTraffic(ctx, 7, duration, stopCh)
+		return generateTrafficAtFixedConcurrency(ctx, 7, duration, stopCh)
 	})
 
 	// Wait for the activator endpoints to equalize.
 	if err := waitForActivatorEndpoints(ctx.resources, ctx.clients); err != nil {
-		t.Fatal("Never got Activator endpoints in the service")
+		t.Fatalf("Never got Activator endpoints in the service: %v", err)
 	}
 
 	// Start second load generator.
 	grp.Go(func() error {
-		return generateTraffic(ctx, 5, duration, stopCh)
+		return generateTrafficAtFixedConcurrency(ctx, 5, duration, stopCh)
 	})
 
 	// Wait for two stable pods.
@@ -574,7 +545,7 @@ func TestTargetBurstCapacityMinusOne(t *testing.T) {
 
 	// Wait for the activator endpoints to equalize.
 	if err := waitForActivatorEndpoints(ctx.resources, ctx.clients); err != nil {
-		t.Fatal("Never got Activator endpoints in the service")
+		t.Fatalf("Never got Activator endpoints in the service: %v", err)
 	}
 }
 
